//
// RoombaComm Serial Interface
//
//  Tod E. Kurt, tod@todbot.com
//
//

import processing.serial.*;

// -------------------------------------------------------------------------
//
// -------------------------------------------------------------------------
class RoombaComm
{
  PApplet parent;
  Serial port;
  String[] portlist;
  boolean connected = false;
  boolean sensorsValid = false;
  boolean safetyFault = false;
  byte[] sensor_bytes = new byte[26];

  boolean debug = true;

  // constructor
  RoombaComm(PApplet p) {
    portlist = Serial.list();
    if(debug) { 
      for(int i=0;i<portlist.length; i++) { // List all the available serial ports:
          println(i+":"+portlist[i]);
      }
    }
    parent = p;
  }
  
  //
  //  returns true on successful connect, false otherwise
  //
  boolean connect(int portnum) {
    String portname = portlist[portnum];
    println("connecting to port '"+portname+"'");
    port = new Serial(parent, portname, 57600);
    connected = port != null && port.port != null; // about the only way I know to do this

    if( connected ) {
      byte cmd[] = { BAUD, 7 };  // shift down to 19200
      port.write(cmd);
      port.dispose();
      port = new Serial(parent, portname, 19200);
      connected = port != null && port.port != null; 
      if( !connected )
        println("connect: couldn't downshift to 19200bps");
    }
    wakeup();
    sensorsValid = false;
    return connected;
  }
  
  //
  // toggles DD line via serial port DTR
  //
  void wakeup() {
    // unimplementable until DTR control is available  
    // can probably do:
    port.port.setDTR(false);
    try { Thread.sleep(500); } catch(Exception e) { }
    port.port.setDTR(true);
  }
  
  //
  //
  //
 void startup() {
    if(debug) println("RoombaComm:startup");
    port.write( START );
    port.write( CONTROL );
    //port.write( FULL );  // dangerous
  }
  
  // FIXME: put this in RoombaComm?
  void reset()
  {
    if(debug) println("RoombaComm:reset");
    startup();
    updateSensors();
  }

  //
  // move the roomba via the low-level velocity + radius method
  //
  void drive( int velocity, int radius ) {
    byte cmd[] = { DRIVE, byte(velocity >>> 8), byte(velocity & 0xff), 
                           byte(radius  >>> 8), byte(radius   & 0xff) };
    if(debug) println("RoombaComm:drive: "+hex(cmd[0])+","+hex(cmd[1])+","+hex(cmd[2])+","+hex(cmd[3])+","+hex(cmd[4]));
    port.write(cmd);
  }
  
  //
  // higher-level movement commands 
  // 
  void stop() {
    if(debug) println("RoombaComm:stop");
    drive( 0, 0 );
  }
  void spinLeft( int speed ) {
    drive( speed, 1 ); 
  }
  void spinRight( int speed ) {
    drive( speed, -1 ); 
  }
  void spinLeft() {
    spinLeft( 100 );  
  }
  void spinRight() {
    spinRight( 100 );  
  }
  void goStraight( int speed ) {
    drive( speed, 0x8000 );
  }
  void goStraight() {
     goStraight( 200 );
  }
  
  // turns on/off the various LEDs
  void setLEDs( boolean status_green, boolean status_red, 
                boolean spot, boolean clean, boolean max, boolean dirt, 
                int power_color, int power_intensity ) {
    int v = (status_green?0x20:0) | (status_red?0x10:0) | (spot?0x08:0) | (clean?0x04:0) | (max?0x02:0) | (dirt?0x01:0);
    println("leds: "+binary(v));
    byte cmd[] = { LEDS, byte(v), byte(power_color), byte(power_intensity) };
    port.write(cmd);
  }
  
  void vacuum( boolean state ) {
    if(debug) println("RoombaComm:vacuum: "+state);
    setMotors( state,state,state );
  }
  
  // turns on/off the non-drive motors
  void setMotors( boolean mainbrush, boolean vacuum, boolean sidebrush ) {
    port.write( MOTORS );
    port.write( byte( (mainbrush?0x04:0) | (vacuum?0x02:0) | (sidebrush?0x01:0) ) );
  }
  
  // plays a note via the hacky method of defining a one-note song & playing it
  // uses up song slot 15
  void playNote( int note, int duration ) {
    if(debug) println("RoombaComm:playnote: "+note+":"+duration);
    byte cmd[] = { SONG, 15, 1, byte(note), byte(duration),
                    PLAY, 15 };
    port.write(cmd);
  }
  
  //
  // returns true on successful sensor update, false otherwise
  //
  boolean updateSensors() {
    byte cmd[] = { SENSORS, 0 };
    port.write(cmd);
    sensorsValid = false;  // assume the worst
    int n = port.available();
    if( n >= 26) {                       // there are enough bytes to read
      n = port.readBytes(sensor_bytes);
      if( n==26 ) {                      // did we get enough?
        sensorsValid = true;             // then everything's good, otherwise bad
        computeSafetyFault();
      }
    } else {
      if(debug) println("RoombaComm:updateSensors:only "+n+" bytes available, not updating sensors");
//      println("buffer contains: "+ port.readString() );
    }
    return sensorsValid;
  }
  
  // if true, indicates we had an event that took the Roomba out of safe mode
  // called on every successful updateSensors
  // use the public boolean 'safety_fault' in your code
  boolean computeSafetyFault() {
    safetyFault = (sensor_bytes[BUMPSWHEELDROPS] & WHEELDROP_MASK) != 0 ||
            sensor_bytes[CLIFFLEFT] == 1  || sensor_bytes[CLIFFFRONTLEFT] == 1 ||
            sensor_bytes[CLIFFRIGHT] == 1  || sensor_bytes[CLIFFFRONTRIGHT] == 1;
    return safetyFault;
  }

  // higher-level sensor access  (FIXME: optimize by accessing sensor_bytes)
  
  boolean bump() {
    return (sensor_bytes[BUMPSWHEELDROPS] & BUMP_MASK) !=0;
  }  
  boolean bumpLeft() {
    return (sensor_bytes[BUMPSWHEELDROPS] & BUMPLEFT_MASK) !=0;
  }
  boolean bumpRight() {
    return (sensor_bytes[BUMPSWHEELDROPS] & BUMPRIGHT_MASK) !=0;
  }
  boolean wheelDropLeft() {
    return (sensor_bytes[BUMPSWHEELDROPS] & WHEELDROPLEFT_MASK) !=0;
  }
  boolean wheelDropRight() {
    return (sensor_bytes[BUMPSWHEELDROPS] & WHEELDROPRIGHT_MASK) !=0;
  }
  boolean wheelDropCenter() {
    return (sensor_bytes[BUMPSWHEELDROPS] & WHEELDROPCENT_MASK) !=0;
  }
  boolean dirt() {
    int dl = sensor_bytes[DIRTLEFT];
    int dr = sensor_bytes[DIRTRIGHT];
    //if(debug) println("RoombaComm:dirt: dl,dr="+dl+","+dr);
    return (dl > 100) || (dr > 100);
  }

  boolean cliffLeft() {
    return (sensor_bytes[CLIFFLEFT] != 0);
  }  
  boolean cliffFrontLeft() {
    return (sensor_bytes[CLIFFFRONTLEFT] != 0);
  }  
  boolean cliffFrontRight() {
    return (sensor_bytes[CLIFFFRONTRIGHT] != 0);
  }  
  boolean cliffRight() {
    return sensor_bytes[CLIFFRIGHT] != 0;
  }
  
  boolean powerButton() {
    return (sensor_bytes[BUTTONS] & POWERBUTTON_MASK) != 0;
  }
  boolean spotButton() {
    return (sensor_bytes[BUTTONS] & SPOTBUTTON_MASK) != 0;
  }
  boolean cleanButton() {
    return (sensor_bytes[BUTTONS] & CLEANBUTTON_MASK) != 0;
  }
  boolean maxButton() {
    return (sensor_bytes[BUTTONS] & MAXBUTTON_MASK) != 0;
  }

  // lower-level sensor access
  
  int bumps_wheeldrops() {
    return sensor_bytes[BUMPSWHEELDROPS];
  }
  int wall() {
    return sensor_bytes[WALL];
  }
  int cliff_left() {
    return sensor_bytes[CLIFFLEFT];
  }
  int cliff_frontleft() {
    return sensor_bytes[CLIFFFRONTLEFT];
  }
  int cliff_frontright() {
    return sensor_bytes[CLIFFFRONTRIGHT];
  }
  int cliff_right() {
    return sensor_bytes[CLIFFRIGHT];
  }
  int virtual_wall() {
    return sensor_bytes[VIRTUALWALL];
  }
  int motor_overcurrents() {
    return sensor_bytes[MOTOROVERCURRENTS];
  }
  int dirt_left() {
    return sensor_bytes[DIRTLEFT];
  }
  int dirt_right() {
    return sensor_bytes[DIRTRIGHT];
  }

  int remote_opcode() {
    return sensor_bytes[REMOTEOPCODE];
  }
  int buttons() {
    return sensor_bytes[BUTTONS];
  }
  int distance() {
    return sensor_bytes[DISTANCE_HI] << 8 | sensor_bytes[DISTANCE_LO];
  }
  int angle() {
    return sensor_bytes[DISTANCE_HI] << 8 | sensor_bytes[ANGLE_LO];
  }
  
  int charging_state() {
    return sensor_bytes[CHARGINGSTATE];
  }
  int voltage() {
    return sensor_bytes[VOLTAGE_HI] << 8 | sensor_bytes[VOLTAGE_LO];
  }
  int current() {
    return sensor_bytes[CURRENT_HI] << 8 | sensor_bytes[CURRENT_LO];
  }
  int temperature() {
    return sensor_bytes[TEMPERATURE];
  }
  int charge() {
    return sensor_bytes[CHARGE_HI] << 8 | sensor_bytes[CHARGE_LO];
  }
  int capacity() {
    return sensor_bytes[CAPACITY_HI] << 8 | sensor_bytes[CAPACITY_LO];
  }

    
  // Roomba SCI opcodes
  // the dumb byte...byte() crap is because processing/java bytes are signed
  byte START   =  byte(128);  // 0
  byte BAUD    =  byte(129);  // 1
  byte CONTROL =  byte(130);  // 0
  byte SAFE    =  byte(131);  // 0
  byte FULL    =  byte(132);  // 0
  byte POWER   =  byte(133);  // 0
  byte SPOT    =  byte(134);  // 0
  byte CLEAN   =  byte(135);  // 0
  byte MAX     =  byte(136);  // 0
  byte DRIVE   =  byte(137);  // 4
  byte MOTORS  =  byte(138);  // 1
  byte LEDS    =  byte(139);  // 3
  byte SONG    =  byte(140);  // 2N+2
  byte PLAY    =  byte(141);  // 1
  byte SENSORS =  byte(142);  // 1
  byte DOCK    =  byte(143);  // 0

  // offsets into sensor_bytes data
  byte BUMPSWHEELDROPS     = 0;
  byte WALL                = 1;
  byte CLIFFLEFT           = 2;
  byte CLIFFFRONTLEFT      = 3;
  byte CLIFFFRONTRIGHT     = 4;
  byte CLIFFRIGHT          = 5;
  byte VIRTUALWALL         = 6;
  byte MOTOROVERCURRENTS   = 7;
  byte DIRTLEFT            = 8;
  byte DIRTRIGHT           = 9;
  byte REMOTEOPCODE        = 10;
  byte BUTTONS             = 11;
  byte DISTANCE_HI         = 12;
  byte DISTANCE_LO         = 13;  
  byte ANGLE_HI            = 14;
  byte ANGLE_LO            = 15;
  byte CHARGINGSTATE       = 16;
  byte VOLTAGE_HI          = 17;
  byte VOLTAGE_LO          = 18;  
  byte CURRENT_HI          = 19;
  byte CURRENT_LO          = 20;
  byte TEMPERATURE         = 21;
  byte CHARGE_HI           = 22;
  byte CHARGE_LO           = 23;
  byte CAPACITY_HI         = 24;
  byte CAPACITY_LO         = 25;

  // bitmasks for various thingems
  byte WHEELDROP_MASK      = 0x1C;
  byte BUMP_MASK           = 0x03;
  byte BUMPRIGHT_MASK      = 0x01;
  byte BUMPLEFT_MASK       = 0x02;
  byte WHEELDROPRIGHT_MASK = 0x04;
  byte WHEELDROPLEFT_MASK  = 0x08;
  byte WHEELDROPCENT_MASK  = 0x10;
  
  byte POWERBUTTON_MASK    = 0x08;  
  byte SPOTBUTTON_MASK     = 0x04;  
  byte CLEANBUTTON_MASK    = 0x02;  
  byte MAXBUTTON_MASK      = 0x01;  

}
